@@ -66,17 +66,30 @@
  */
 #define TOUCH_AXIS_MAX 0xffff
 
+#define CAP_KEYBOARD	0x1
+#define CAP_POINTER	0x2
+#define CAP_TOUCH	0x4
+
 struct xf86libinput_driver {
 	struct libinput *libinput;
 	int device_enabled_count;
-	struct xorg_list server_fds;
 };
 
 static struct xf86libinput_driver driver_context;
 
-struct xf86libinput {
-	char *path;
+struct xf86libinput_device {
+	int refcount;
+	int enabled_count;
+	uint32_t id;
 	struct libinput_device *device;
+	struct xorg_list device_list;
+	int server_fd;
+};
+
+struct xf86libinput {
+	InputInfoPtr pInfo;
+	char *path;
+	uint32_t capabilities;
 
 	struct {
 		int vdist;
@@ -99,6 +112,7 @@ struct xf86libinput {
 
 	struct options {
 		BOOL tapping;
+		BOOL tap_drag;
 		BOOL tap_drag_lock;
 		BOOL natural_scrolling;
 		BOOL left_handed;
@@ -118,6 +132,14 @@ struct xf86libinput {
 	} options;
 
 	struct draglock draglock;
+
+	struct xf86libinput_device *shared_device;
+	struct xorg_list shared_device_link;
+};
+
+enum hotplug_when {
+	HOTPLUG_LATER,
+	HOTPLUG_NOW,
 };
 
 static inline int
@@ -160,6 +182,141 @@ btn_xorg2linux(unsigned int b)
 	return button;
 }
 
+static inline InputInfoPtr
+xf86libinput_get_parent(InputInfoPtr pInfo)
+{
+	InputInfoPtr parent;
+	int parent_id;
+
+	parent_id = xf86CheckIntOption(pInfo->options, "_libinput/shared-device", -1);
+	if (parent_id == -1)
+		return NULL;
+
+	nt_list_for_each_entry(parent, xf86FirstLocalDevice(), next) {
+		int id = xf86CheckIntOption(parent->options,
+					    "_libinput/shared-device",
+					    -1);
+		if (id == parent_id)
+			return parent;
+	}
+
+	return NULL;
+}
+
+static inline struct xf86libinput_device*
+xf86libinput_shared_create(struct libinput_device *device)
+{
+	static uint32_t next_shared_device_id;
+	struct xf86libinput_device *shared_device;
+
+	shared_device = calloc(1, sizeof(*shared_device));
+	if (!shared_device)
+		return NULL;
+
+	shared_device->device = device;
+	shared_device->refcount = 1;
+	shared_device->id = ++next_shared_device_id;
+	xorg_list_init(&shared_device->device_list);
+
+	return shared_device;
+}
+
+static inline struct xf86libinput_device*
+xf86libinput_shared_ref(struct xf86libinput_device *shared_device)
+{
+	shared_device->refcount++;
+
+	return shared_device;
+}
+
+static inline struct xf86libinput_device*
+xf86libinput_shared_unref(struct xf86libinput_device *shared_device)
+{
+	shared_device->refcount--;
+
+	if (shared_device->refcount > 0)
+		return shared_device;
+
+	free(shared_device);
+
+	return NULL;
+}
+
+static inline struct libinput_device *
+xf86libinput_shared_enable(InputInfoPtr pInfo,
+			   struct xf86libinput_device *shared_device,
+			   const char *path)
+{
+	struct libinput_device *device;
+	struct libinput *libinput = driver_context.libinput;
+
+	/* With systemd-logind the server requests the fd from logind, sets
+	 * pInfo->fd and sets the "fd" option to the fd number.
+	 *
+	 * If we have a second device that uses the same path, the server
+	 * checks all pInfo->major/minor for a match and returns the matched
+	 * device's pInfo->fd. In this driver, this fd is the epollfd, not
+	 * the actual device. This causes troubles when removing the
+	 * device.
+	 *
+	 * What we need to do here is: after enabling the device the first
+	 * time extract the real fd and store it in the shared device
+	 * struct. The second device replaces the pInfo->options "fd" with
+	 * the real fd we're using.
+	 *
+	 * When the device is unplugged, the server now correctly finds two
+	 * devices on the real fd and releases them in order.
+	 */
+	shared_device->enabled_count++;
+	if (shared_device->enabled_count > 1) {
+		if (pInfo->flags & XI86_SERVER_FD) {
+			pInfo->options = xf86ReplaceIntOption(pInfo->options,
+							      "fd",
+							      shared_device->server_fd);
+		}
+
+		return shared_device->device;
+	}
+
+	device = libinput_path_add_device(libinput, path);
+	if (!device)
+		return NULL;
+
+	libinput_device_set_user_data(device, shared_device);
+	shared_device->device = libinput_device_ref(device);
+
+	if (pInfo->flags & XI86_SERVER_FD)
+		shared_device->server_fd = xf86CheckIntOption(pInfo->options,
+							      "fd",
+							      -1);
+	return device;
+}
+
+static inline void
+xf86libinput_shared_disable(struct xf86libinput_device *shared_device)
+{
+	struct libinput_device *device = shared_device->device;
+
+	shared_device->enabled_count--;
+
+	if (shared_device->enabled_count > 0)
+		return;
+
+	if (!device)
+		return;
+
+	libinput_device_set_user_data(device, NULL);
+	libinput_path_remove_device(device);
+	device = libinput_device_unref(device);
+	shared_device->device = NULL;
+}
+
+static inline bool
+xf86libinput_shared_is_enabled(struct xf86libinput_device *shared_device)
+{
+	return shared_device->enabled_count > 0;
+}
+
 static int
 LibinputSetProperty(DeviceIntPtr dev, Atom atom, XIPropertyValuePtr val,
                  BOOL checkonly);
@@ -171,7 +328,7 @@ LibinputApplyConfig(DeviceIntPtr dev)
 {
 	InputInfoPtr pInfo = dev->public.devicePrivate;
 	struct xf86libinput *driver_data = pInfo->private;
-	struct libinput_device *device = driver_data->device;
+	struct libinput_device *device = driver_data->shared_device->device;
 	unsigned int scroll_button;
 
 	if (libinput_device_config_send_events_get_modes(device) != LIBINPUT_CONFIG_SEND_EVENTS_ENABLED &&
@@ -230,6 +387,13 @@ LibinputApplyConfig(DeviceIntPtr dev)
 			    "Failed to set Tapping DragLock to %d\n",
 			    driver_data->options.tap_drag_lock);
 
+	if (libinput_device_config_tap_get_finger_count(device) > 0 &&
+	    libinput_device_config_tap_set_drag_enabled(device,
+							driver_data->options.tap_drag) != LIBINPUT_CONFIG_STATUS_SUCCESS)
+		xf86IDrvMsg(pInfo, X_ERROR,
+			    "Failed to set Tapping Drag to %d\n",
+			    driver_data->options.tap_drag);
+
 	if (libinput_device_config_calibration_has_matrix(device) &&
 	    libinput_device_config_calibration_set_matrix(device,
 							  driver_data->options.matrix) != LIBINPUT_CONFIG_STATUS_SUCCESS)
@@ -312,17 +476,16 @@ xf86libinput_on(DeviceIntPtr dev)
 {
 	InputInfoPtr pInfo = dev->public.devicePrivate;
 	struct xf86libinput *driver_data = pInfo->private;
+	struct xf86libinput_device *shared_device = driver_data->shared_device;
 	struct libinput *libinput = driver_context.libinput;
-	struct libinput_device *device = driver_data->device;
+	struct libinput_device *device;
 
-	device = libinput_path_add_device(libinput, driver_data->path);
+	device = xf86libinput_shared_enable(pInfo,
+					    shared_device,
+					    driver_data->path);
 	if (!device)
 		return !Success;
 
-	libinput_device_ref(device);
-	libinput_device_set_user_data(device, pInfo);
-	driver_data->device = device;
-
 	/* if we use server fds, overwrite the fd with the one from
 	   libinput nonetheless, otherwise the server won't call ReadInput
 	   for our device. This must be swapped back to the real fd in
@@ -347,6 +510,7 @@ xf86libinput_off(DeviceIntPtr dev)
 {
 	InputInfoPtr pInfo = dev->public.devicePrivate;
 	struct xf86libinput *driver_data = pInfo->private;
+	struct xf86libinput_device *shared_device = driver_data->shared_device;
 
 	if (--driver_context.device_enabled_count == 0) {
 		RemoveEnabledDevice(pInfo->fd);
@@ -360,10 +524,7 @@ xf86libinput_off(DeviceIntPtr dev)
 
 	dev->public.on = FALSE;
 
-	libinput_device_set_user_data(driver_data->device, NULL);
-	libinput_path_remove_device(driver_data->device);
-	libinput_device_unref(driver_data->device);
-	driver_data->device = NULL;
+	xf86libinput_shared_disable(shared_device);
 
 	return Success;
 }
@@ -417,6 +578,7 @@ xf86libinput_init_pointer(InputInfoPtr pInfo)
 {
 	DeviceIntPtr dev= pInfo->dev;
 	struct xf86libinput *driver_data = pInfo->private;
+	struct libinput_device *device = driver_data->shared_device->device;
 	int min, max, res;
 	int nbuttons = 7;
 	int i;
@@ -425,7 +587,7 @@ xf86libinput_init_pointer(InputInfoPtr pInfo)
 	Atom axislabels[TOUCHPAD_NUM_AXES];
 
 	for (i = BTN_JOYSTICK - 1; i >= BTN_SIDE; i--) {
-		if (libinput_device_pointer_has_button(driver_data->device, i)) {
+		if (libinput_device_pointer_has_button(device, i)) {
 			nbuttons += i - BTN_SIDE + 1;
 			break;
 		}
@@ -464,6 +626,7 @@ xf86libinput_init_pointer_absolute(InputInfoPtr pInfo)
 {
 	DeviceIntPtr dev= pInfo->dev;
 	struct xf86libinput *driver_data = pInfo->private;
+	struct libinput_device *device = driver_data->shared_device->device;
 	int min, max, res;
 	int nbuttons = 7;
 	int i;
@@ -472,7 +635,7 @@ xf86libinput_init_pointer_absolute(InputInfoPtr pInfo)
 	Atom axislabels[TOUCHPAD_NUM_AXES];
 
 	for (i = BTN_BACK; i >= BTN_SIDE; i--) {
-		if (libinput_device_pointer_has_button(driver_data->device, i)) {
+		if (libinput_device_pointer_has_button(device, i)) {
 			nbuttons += i - BTN_SIDE + 1;
 			break;
 		}
@@ -524,7 +687,7 @@ xf86libinput_kbd_ctrl(DeviceIntPtr device, KeybdCtrl *ctrl)
     enum libinput_led leds = 0;
     InputInfoPtr pInfo = device->public.devicePrivate;
     struct xf86libinput *driver_data = pInfo->private;
-    struct libinput_device *ldevice = driver_data->device;
+    struct libinput_device *ldevice = driver_data->shared_device->device;
 
     while (bits[i].xbit) {
 	    if (ctrl->leds & bits[i].xbit)
@@ -608,29 +771,37 @@ xf86libinput_init(DeviceIntPtr dev)
 {
 	InputInfoPtr pInfo = dev->public.devicePrivate;
 	struct xf86libinput *driver_data = pInfo->private;
-	struct libinput_device *device = driver_data->device;
+	struct xf86libinput_device *shared_device = driver_data->shared_device;
+	struct libinput_device *device = shared_device->device;
+
+	BUG_RETURN_VAL(device == NULL, !Success);
 
 	dev->public.on = FALSE;
 
-	if (libinput_device_has_capability(device, LIBINPUT_DEVICE_CAP_KEYBOARD))
+	if (driver_data->capabilities & CAP_KEYBOARD)
 		xf86libinput_init_keyboard(pInfo);
-	if (libinput_device_has_capability(device, LIBINPUT_DEVICE_CAP_POINTER)) {
+	if (driver_data->capabilities & CAP_POINTER) {
 		if (libinput_device_config_calibration_has_matrix(device) &&
 		    !libinput_device_config_accel_is_available(device))
 			xf86libinput_init_pointer_absolute(pInfo);
 		else
 			xf86libinput_init_pointer(pInfo);
 	}
-	if (libinput_device_has_capability(device, LIBINPUT_DEVICE_CAP_TOUCH))
+	if (driver_data->capabilities & CAP_TOUCH)
 		xf86libinput_init_touch(pInfo);
 
 	LibinputApplyConfig(dev);
 	LibinputInitProperty(dev);
 	XIRegisterPropertyHandler(dev, LibinputSetProperty, NULL, NULL);
 
-	/* unref the device now, because we'll get a new ref during
-	   DEVICE_ON */
-	libinput_device_unref(device);
+	/* If we have a device but it's not yet enabled it's the
+	 * already-removed device from PreInit. Drop the ref to clean up,
+	 * we'll get a new libinput_device during DEVICE_ON when we re-add
+	 * it. */
+	if (!xf86libinput_shared_is_enabled(shared_device)) {
+		libinput_device_unref(device);
+		shared_device->device = NULL;
+	}
 
 	return 0;
 }
@@ -638,6 +809,13 @@ xf86libinput_init(DeviceIntPtr dev)
 static void
 xf86libinput_destroy(DeviceIntPtr dev)
 {
+	InputInfoPtr pInfo = dev->public.devicePrivate;
+	struct xf86libinput *driver_data = pInfo->private;
+	struct xf86libinput_device *shared_device = driver_data->shared_device;
+
+	xorg_list_del(&driver_data->shared_device_link);
+
+	xf86libinput_shared_unref(shared_device);
 }
 
 static int
@@ -657,6 +835,7 @@ xf86libinput_device_control(DeviceIntPtr dev, int mode)
 			break;
 		case DEVICE_CLOSE:
 			xf86libinput_destroy(dev);
+			rc = Success;
 			break;
 	}
 
@@ -671,6 +850,9 @@ xf86libinput_handle_motion(InputInfoPtr pInfo, struct libinput_event_pointer *ev
 	ValuatorMask *mask = driver_data->valuators;
 	double x, y;
 
+	if ((driver_data->capabilities & CAP_POINTER) == 0)
+		return;
+
 	x = libinput_event_pointer_get_dx(event);
 	y = libinput_event_pointer_get_dy(event);
 
@@ -708,6 +890,9 @@ xf86libinput_handle_absmotion(InputInfoPtr pInfo, struct libinput_event_pointer
 		return;
 	}
 
+	if ((driver_data->capabilities & CAP_POINTER) == 0)
+		return;
+
 	x = libinput_event_pointer_get_absolute_x_transformed(event, TOUCH_AXIS_MAX);
 	y = libinput_event_pointer_get_absolute_y_transformed(event, TOUCH_AXIS_MAX);
 
@@ -726,6 +911,9 @@ xf86libinput_handle_button(InputInfoPtr pInfo, struct libinput_event_pointer *ev
 	int button;
 	int is_press;
 
+	if ((driver_data->capabilities & CAP_POINTER) == 0)
+		return;
+
 	button = btn_linux2xorg(libinput_event_pointer_get_button(event));
 	is_press = (libinput_event_pointer_get_button_state(event) == LIBINPUT_BUTTON_STATE_PRESSED);
 
@@ -740,9 +928,13 @@ static void
 xf86libinput_handle_key(InputInfoPtr pInfo, struct libinput_event_keyboard *event)
 {
 	DeviceIntPtr dev = pInfo->dev;
+	struct xf86libinput *driver_data = pInfo->private;
 	int is_press;
 	int key = libinput_event_keyboard_get_key(event);
 
+	if ((driver_data->capabilities & CAP_KEYBOARD) == 0)
+		return;
+
 	key += XORG_KEYCODE_OFFSET;
 
 	is_press = (libinput_event_keyboard_get_key_state(event) == LIBINPUT_KEY_STATE_PRESSED);
@@ -759,6 +951,9 @@ xf86libinput_handle_axis(InputInfoPtr pInfo, struct libinput_event_pointer *even
 	enum libinput_pointer_axis axis;
 	enum libinput_pointer_axis_source source;
 
+	if ((driver_data->capabilities & CAP_POINTER) == 0)
+		return;
+
 	valuator_mask_zero(mask);
 
 	source = libinput_event_pointer_get_axis_source(event);
@@ -817,6 +1012,9 @@ xf86libinput_handle_touch(InputInfoPtr pInfo,
 	static unsigned int next_touchid;
 	static unsigned int touchids[TOUCH_MAX_SLOTS] = {0};
 
+	if ((driver_data->capabilities & CAP_TOUCH) == 0)
+		return;
+
 	slot = libinput_event_touch_get_slot(event);
 
 	switch (event_type) {
@@ -847,19 +1045,50 @@ xf86libinput_handle_touch(InputInfoPtr pInfo,
 	xf86PostTouchEvent(dev, touchids[slot], type, 0, m);
 }
 
+static InputInfoPtr
+xf86libinput_pick_device(struct xf86libinput_device *shared_device,
+			 enum libinput_event_type type)
+{
+	struct xf86libinput *driver_data;
+	uint32_t needed_cap;
+
+	if (shared_device == NULL)
+		return NULL;
+
+	switch(type) {
+	case LIBINPUT_EVENT_KEYBOARD_KEY:
+		needed_cap = CAP_KEYBOARD;
+		break;
+	default:
+		needed_cap = ~CAP_KEYBOARD;
+		break;
+	}
+
+	xorg_list_for_each_entry(driver_data,
+				 &shared_device->device_list,
+				 shared_device_link) {
+		if (driver_data->capabilities & needed_cap)
+			return driver_data->pInfo;
+	}
+
+	return NULL;
+}
+
 static void
 xf86libinput_handle_event(struct libinput_event *event)
 {
 	struct libinput_device *device;
+	enum libinput_event_type type;
 	InputInfoPtr pInfo;
 
+	type = libinput_event_get_type(event);
 	device = libinput_event_get_device(event);
-	pInfo = libinput_device_get_user_data(device);
+	pInfo = xf86libinput_pick_device(libinput_device_get_user_data(device), type);
 
 	if (!pInfo || !pInfo->dev->public.on)
 		return;
 
-	switch (libinput_event_get_type(event)) {
+	switch (type) {
 		case LIBINPUT_EVENT_NONE:
 		case LIBINPUT_EVENT_DEVICE_ADDED:
 		case LIBINPUT_EVENT_DEVICE_REMOVED:
@@ -902,6 +1131,11 @@ xf86libinput_handle_event(struct libinput_event *event)
 		case LIBINPUT_EVENT_GESTURE_PINCH_UPDATE:
 		case LIBINPUT_EVENT_GESTURE_PINCH_END:
 			break;
+		case LIBINPUT_EVENT_TABLET_TOOL_AXIS:
+		case LIBINPUT_EVENT_TABLET_TOOL_BUTTON:
+		case LIBINPUT_EVENT_TABLET_TOOL_PROXIMITY:
+		case LIBINPUT_EVENT_TABLET_TOOL_TIP:
+			break;
 	}
 }
 
@@ -959,7 +1193,12 @@ open_restricted(const char *path, int flags, void *data)
 	}
 
 	fd = xf86OpenSerial(pInfo->options);
-	return fd < 0 ? -errno : fd;
+	if (fd < 0)
+		return -errno;
+
+	xf86FlushInput(fd);
+
+	return fd;
 }
 
 static void
@@ -1050,6 +1289,30 @@ xf86libinput_parse_tap_option(InputInfoPtr pInfo,
 	return tap;
 }
 
+static inline BOOL
+xf86libinput_parse_tap_drag_option(InputInfoPtr pInfo,
+				   struct libinput_device *device)
+{
+	BOOL drag;
+
+	if (libinput_device_config_tap_get_finger_count(device) == 0)
+		return FALSE;
+
+	drag = xf86SetBoolOption(pInfo->options,
+				 "TappingDrag",
+				 libinput_device_config_tap_get_drag_enabled(device));
+
+	if (libinput_device_config_tap_set_drag_enabled(device, drag) !=
+	    LIBINPUT_CONFIG_STATUS_SUCCESS) {
+		xf86IDrvMsg(pInfo, X_ERROR,
+			    "Failed to set Tapping Drag Lock to %d\n",
+			    drag);
+		drag = libinput_device_config_tap_get_drag_enabled(device);
+	}
+
+	return drag;
+}
+
 static inline BOOL
 xf86libinput_parse_tap_drag_lock_option(InputInfoPtr pInfo,
 					struct libinput_device *device)
@@ -1122,6 +1385,8 @@ xf86libinput_parse_accel_profile_option(InputInfoPtr pInfo,
 		profile = libinput_device_config_accel_get_profile(device);
 	}
 
+	free(str);
+
 	return profile;
 }
 
@@ -1455,6 +1720,7 @@ xf86libinput_parse_options(InputInfoPtr pInfo,
 
 	/* libinput options */
 	options->tapping = xf86libinput_parse_tap_option(pInfo, device);
+	options->tap_drag = xf86libinput_parse_tap_drag_option(pInfo, device);
 	options->tap_drag_lock = xf86libinput_parse_tap_drag_lock_option(pInfo, device);
 	options->speed = xf86libinput_parse_accel_option(pInfo, device);
 	options->accel_profile = xf86libinput_parse_accel_profile_option(pInfo, device);
@@ -1472,21 +1738,171 @@ xf86libinput_parse_options(InputInfoPtr pInfo,
 	xf86libinput_parse_buttonmap_option(pInfo,
 					    options->btnmap,
 					    sizeof(options->btnmap));
-	if (libinput_device_has_capability(device, LIBINPUT_DEVICE_CAP_POINTER)) {
+	if (driver_data->capabilities & CAP_POINTER) {
 		xf86libinput_parse_draglock_option(pInfo, driver_data);
 		options->horiz_scrolling_enabled = xf86libinput_parse_horiz_scroll_option(pInfo);
 	}
 }
 
+static const char*
+xf86libinput_get_type_name(struct libinput_device *device,
+			   struct xf86libinput *driver_data)
+{
+	const char *type_name;
+
+	/* now pick an actual type */
+	if (libinput_device_config_tap_get_finger_count(device) > 0)
+		type_name = XI_TOUCHPAD;
+	else if (driver_data->capabilities & CAP_TOUCH)
+		type_name = XI_TOUCHSCREEN;
+	else if (driver_data->capabilities & CAP_POINTER)
+		type_name = XI_MOUSE;
+	else
+		type_name = XI_KEYBOARD;
+
+	return type_name;
+}
+
+static void
+xf86libinput_init_driver_context(void)
+{
+	if (!driver_context.libinput) {
+		driver_context.libinput = libinput_path_create_context(&interface, &driver_context);
+		libinput_log_set_handler(driver_context.libinput,
+					 xf86libinput_log_handler);
+		/* we want all msgs, let the server filter */
+		libinput_log_set_priority(driver_context.libinput,
+					  LIBINPUT_LOG_PRIORITY_DEBUG);
+	} else {
+		libinput_ref(driver_context.libinput);
+	}
+}
+
+struct xf86libinput_hotplug_info {
+	InputAttributes *attrs;
+	InputOption *input_options;
+};
+
+static DeviceIntPtr
+xf86libinput_hotplug_device(struct xf86libinput_hotplug_info *hotplug)
+{
+	DeviceIntPtr dev;
+
+	if (NewInputDeviceRequest(hotplug->input_options,
+				  hotplug->attrs,
+				  &dev) != Success)
+		dev = NULL;
+
+	input_option_free_list(&hotplug->input_options);
+	FreeInputAttributes(hotplug->attrs);
+	free(hotplug);
+
+	return dev;
+}
+
+static Bool
+xf86libinput_hotplug_device_cb(ClientPtr client, pointer closure)
+{
+	struct xf86libinput_hotplug_info *hotplug = closure;
+
+	xf86libinput_hotplug_device(hotplug);
+
+	return TRUE;
+}
+
+static DeviceIntPtr
+xf86libinput_create_subdevice(InputInfoPtr pInfo,
+			      uint32_t capabilities,
+			      enum hotplug_when when,
+			      XF86OptionPtr extra_options)
+{
+	struct xf86libinput *driver_data = pInfo->private;
+	struct xf86libinput_device *shared_device;
+	struct xf86libinput_hotplug_info *hotplug;
+	InputOption *iopts = NULL;
+	XF86OptionPtr options, o;
+
+	shared_device = driver_data->shared_device;
+	pInfo->options = xf86ReplaceIntOption(pInfo->options,
+					      "_libinput/shared-device",
+					      shared_device->id);
+
+	options = xf86OptionListDuplicate(pInfo->options);
+	options = xf86ReplaceStrOption(options, "_source", "_driver/libinput");
+	options = xf86OptionListMerge(options, extra_options);
+
+	if (capabilities & CAP_KEYBOARD)
+		options = xf86ReplaceBoolOption(options, "_libinput/cap-keyboard", 1);
+	if (capabilities & CAP_POINTER)
+		options = xf86ReplaceBoolOption(options, "_libinput/cap-pointer", 1);
+	if (capabilities & CAP_TOUCH)
+		options = xf86ReplaceBoolOption(options, "_libinput/cap-touch", 1);
+
+	/* need convert from one option list to the other. woohoo. */
+	o = options;
+	while (o) {
+		iopts = input_option_new(iopts,
+					 xf86OptionName(o),
+					 xf86OptionValue(o));
+		o = xf86NextOption(o);
+	}
+	xf86OptionListFree(options);
+
+	hotplug = calloc(1, sizeof(*hotplug));
+	if (!hotplug)
+		return NULL;
+
+	hotplug->input_options = iopts;
+	hotplug->attrs = DuplicateInputAttributes(pInfo->attrs);
+
+	xf86IDrvMsg(pInfo, X_INFO, "needs a virtual subdevice\n");
+	if (when == HOTPLUG_LATER)
+		QueueWorkProc(xf86libinput_hotplug_device_cb, serverClient, hotplug);
+	else
+		return xf86libinput_hotplug_device(hotplug);
+
+	return NULL;
+}
+
+static BOOL
+xf86libinput_is_subdevice(InputInfoPtr pInfo)
+{
+	char *source;
+	BOOL is_subdevice;
+
+	source = xf86SetStrOption(pInfo->options, "_source", "");
+	is_subdevice = strcmp(source, "_driver/libinput") == 0;
+	free(source);
+
+	return is_subdevice;
+}
+
+static inline uint32_t
+caps_from_options(InputInfoPtr pInfo)
+{
+	uint32_t capabilities = 0;
+
+	if (xf86CheckBoolOption(pInfo->options, "_libinput/cap-keyboard", 0))
+		capabilities |= CAP_KEYBOARD;
+	if (xf86CheckBoolOption(pInfo->options, "_libinput/cap-pointer", 0))
+		capabilities |= CAP_POINTER;
+	if (xf86CheckBoolOption(pInfo->options, "_libinput/cap-touch", 0))
+		capabilities |= CAP_TOUCH;
+
+	return capabilities;
+}
+
 static int
 xf86libinput_pre_init(InputDriverPtr drv,
 		      InputInfoPtr pInfo,
 		      int flags)
 {
 	struct xf86libinput *driver_data = NULL;
-        struct libinput *libinput = NULL;
+	struct xf86libinput_device *shared_device = NULL;
+	struct libinput *libinput = NULL;
 	struct libinput_device *device;
 	char *path = NULL;
+	bool is_subdevice;
 
 	pInfo->type_name = 0;
 	pInfo->device_control = xf86libinput_device_control;
@@ -1506,25 +1922,11 @@ xf86libinput_pre_init(InputDriverPtr drv,
 	if (!driver_data->valuators_unaccelerated)
 		goto fail;
 
-	driver_data->scroll.vdist = 15;
-	driver_data->scroll.hdist = 15;
-
 	path = xf86SetStrOption(pInfo->options, "Device", NULL);
 	if (!path)
 		goto fail;
 
-	if (!driver_context.libinput) {
-		driver_context.libinput = libinput_path_create_context(&interface, &driver_context);
-		libinput_log_set_handler(driver_context.libinput,
-					 xf86libinput_log_handler);
-		/* we want all msgs, let the server filter */
-		libinput_log_set_priority(driver_context.libinput,
-					  LIBINPUT_LOG_PRIORITY_DEBUG);
-		xorg_list_init(&driver_context.server_fds);
-	} else {
-		libinput_ref(driver_context.libinput);
-	}
-
+	xf86libinput_init_driver_context();
 	libinput = driver_context.libinput;
 
 	if (libinput == NULL) {
@@ -1532,21 +1934,60 @@ xf86libinput_pre_init(InputDriverPtr drv,
 		goto fail;
 	}
 
-	device = libinput_path_add_device(libinput, path);
-	if (!device) {
-		xf86IDrvMsg(pInfo, X_ERROR, "Failed to create a device for %s\n", path);
-		goto fail;
+	is_subdevice = xf86libinput_is_subdevice(pInfo);
+	if (!is_subdevice) {
+		device = libinput_path_add_device(libinput, path);
+		if (!device) {
+			xf86IDrvMsg(pInfo, X_ERROR, "Failed to create a device for %s\n", path);
+			goto fail;
+		}
+
+		/* We ref the device above, then remove it. It get's
+		   re-added with the same path in DEVICE_ON, we hope
+		   it doesn't change until then */
+		libinput_device_ref(device);
+		libinput_path_remove_device(device);
+
+		shared_device = xf86libinput_shared_create(device);
+		if (!shared_device) {
+			libinput_device_unref(device);
+			goto fail;
+		}
+	} else {
+		InputInfoPtr parent;
+		struct xf86libinput *parent_driver_data;
+
+		parent = xf86libinput_get_parent(pInfo);
+		if (!parent) {
+			xf86IDrvMsg(pInfo, X_ERROR, "Failed to find parent device\n");
+			goto fail;
+		}
+		xf86IDrvMsg(pInfo, X_INFO, "is a virtual subdevice\n");
+
+		parent_driver_data = parent->private;
+		shared_device = xf86libinput_shared_ref(parent_driver_data->shared_device);
+		device = shared_device->device;
 	}
 
-	/* We ref the device but remove it afterwards. The hope is that
-	   between now and DEVICE_INIT/DEVICE_ON, the device doesn't change.
-	  */
-	libinput_device_ref(device);
-	libinput_path_remove_device(device);
-
 	pInfo->private = driver_data;
+	driver_data->pInfo = pInfo;
+	driver_data->scroll.vdist = 15;
+	driver_data->scroll.hdist = 15;
 	driver_data->path = path;
-	driver_data->device = device;
+	driver_data->shared_device = shared_device;
+	xorg_list_append(&driver_data->shared_device_link,
+			 &shared_device->device_list);
+
+	if (!is_subdevice) {
+		if (libinput_device_has_capability(device, LIBINPUT_DEVICE_CAP_POINTER))
+			driver_data->capabilities |= CAP_POINTER;
+		if (libinput_device_has_capability(device, LIBINPUT_DEVICE_CAP_KEYBOARD))
+			driver_data->capabilities |= CAP_KEYBOARD;
+		if (libinput_device_has_capability(device, LIBINPUT_DEVICE_CAP_TOUCH))
+			driver_data->capabilities |= CAP_TOUCH;
+	} else {
+		driver_data->capabilities = caps_from_options(pInfo);
+	}
 
 	/* Disable acceleration in the server, libinput does it for us */
 	pInfo->options = xf86ReplaceIntOption(pInfo->options, "AccelerationProfile", -1);
@@ -1554,17 +1995,19 @@ xf86libinput_pre_init(InputDriverPtr drv,
 
 	xf86libinput_parse_options(pInfo, driver_data, device);
 
-	/* now pick an actual type */
-	if (libinput_device_config_tap_get_finger_count(device) > 0)
-		pInfo->type_name = XI_TOUCHPAD;
-	else if (libinput_device_has_capability(device,
-						LIBINPUT_DEVICE_CAP_TOUCH))
-		pInfo->type_name = XI_TOUCHSCREEN;
-	else if (libinput_device_has_capability(device,
-						LIBINPUT_DEVICE_CAP_POINTER))
-		pInfo->type_name = XI_MOUSE;
-	else
-		pInfo->type_name = XI_KEYBOARD;
+	/* Device is both keyboard and pointer. Drop the keyboard cap from
+	 * this device, create a separate device instead */
+	if (!is_subdevice &&
+	    driver_data->capabilities & CAP_KEYBOARD &&
+	    driver_data->capabilities & (CAP_POINTER|CAP_TOUCH)) {
+		driver_data->capabilities &= ~CAP_KEYBOARD;
+		xf86libinput_create_subdevice(pInfo,
+					      CAP_KEYBOARD,
+					      HOTPLUG_LATER,
+					      NULL);
+	}
+
+	pInfo->type_name = xf86libinput_get_type_name(device, driver_data);
 
 	return Success;
 fail:
@@ -1573,7 +2016,11 @@ fail:
 	if (driver_data->valuators_unaccelerated)
 		valuator_mask_free(&driver_data->valuators_unaccelerated);
 	free(path);
+	if (shared_device)
+		xf86libinput_shared_unref(shared_device);
 	free(driver_data);
+	if (libinput)
+		driver_context.libinput = libinput_unref(libinput);
 	return BadValue;
 }
 
@@ -1586,6 +2033,7 @@ xf86libinput_uninit(InputDriverPtr drv,
 	if (driver_data) {
 		driver_context.libinput = libinput_unref(driver_context.libinput);
 		valuator_mask_free(&driver_data->valuators);
+		valuator_mask_free(&driver_data->valuators_unaccelerated);
 		free(driver_data->path);
 		free(driver_data);
 		pInfo->private = NULL;
@@ -1636,6 +2084,8 @@ _X_EXPORT XF86ModuleData libinputModuleData = {
 /* libinput-specific properties */
 static Atom prop_tap;
 static Atom prop_tap_default;
+static Atom prop_tap_drag;
+static Atom prop_tap_drag_default;
 static Atom prop_tap_drag_lock;
 static Atom prop_tap_drag_lock_default;
 static Atom prop_calibration;
@@ -1680,7 +2130,7 @@ xf86libinput_check_device (DeviceIntPtr dev,
 {
 	InputInfoPtr pInfo = dev->public.devicePrivate;
 	struct xf86libinput *driver_data = pInfo->private;
-	struct libinput_device *device = driver_data->device;
+	struct libinput_device *device = driver_data->shared_device->device;
 
 	if (device == NULL) {
 		BUG_WARN(dev->public.on);
@@ -1702,7 +2152,7 @@ LibinputSetPropertyTap(DeviceIntPtr dev,
 {
 	InputInfoPtr pInfo = dev->public.devicePrivate;
 	struct xf86libinput *driver_data = pInfo->private;
-	struct libinput_device *device = driver_data->device;
+	struct libinput_device *device = driver_data->shared_device->device;
 	BOOL* data;
 
 	if (val->format != 8 || val->size != 1 || val->type != XA_INTEGER)
@@ -1725,6 +2175,37 @@ LibinputSetPropertyTap(DeviceIntPtr dev,
 	return Success;
 }
 
+static inline int
+LibinputSetPropertyTapDrag(DeviceIntPtr dev,
+			   Atom atom,
+			   XIPropertyValuePtr val,
+			   BOOL checkonly)
+{
+	InputInfoPtr pInfo = dev->public.devicePrivate;
+	struct xf86libinput *driver_data = pInfo->private;
+	struct libinput_device *device = driver_data->shared_device->device;
+	BOOL* data;
+
+	if (val->format != 8 || val->size != 1 || val->type != XA_INTEGER)
+		return BadMatch;
+
+	data = (BOOL*)val->data;
+	if (checkonly) {
+		if (*data != 0 && *data != 1)
+			return BadValue;
+
+		if (!xf86libinput_check_device(dev, atom))
+			return BadMatch;
+
+		if (libinput_device_config_tap_get_finger_count(device) == 0)
+			return BadMatch;
+	} else {
+		driver_data->options.tap_drag = *data;
+	}
+
+	return Success;
+}
+
 static inline int
 LibinputSetPropertyTapDragLock(DeviceIntPtr dev,
 			       Atom atom,
@@ -1733,7 +2214,7 @@ LibinputSetPropertyTapDragLock(DeviceIntPtr dev,
 {
 	InputInfoPtr pInfo = dev->public.devicePrivate;
 	struct xf86libinput *driver_data = pInfo->private;
-	struct libinput_device *device = driver_data->device;
+	struct libinput_device *device = driver_data->shared_device->device;
 	BOOL* data;
 
 	if (val->format != 8 || val->size != 1 || val->type != XA_INTEGER)
@@ -1764,7 +2245,7 @@ LibinputSetPropertyCalibration(DeviceIntPtr dev,
 {
 	InputInfoPtr pInfo = dev->public.devicePrivate;
 	struct xf86libinput *driver_data = pInfo->private;
-	struct libinput_device *device = driver_data->device;
+	struct libinput_device *device = driver_data->shared_device->device;
 	float* data;
 
 	if (val->format != 32 || val->size != 9 || val->type != prop_float)
@@ -1800,7 +2281,7 @@ LibinputSetPropertyAccel(DeviceIntPtr dev,
 {
 	InputInfoPtr pInfo = dev->public.devicePrivate;
 	struct xf86libinput *driver_data = pInfo->private;
-	struct libinput_device *device = driver_data->device;
+	struct libinput_device *device = driver_data->shared_device->device;
 	float* data;
 
 	if (val->format != 32 || val->size != 1 || val->type != prop_float)
@@ -1832,7 +2313,7 @@ LibinputSetPropertyAccelProfile(DeviceIntPtr dev,
 {
 	InputInfoPtr pInfo = dev->public.devicePrivate;
 	struct xf86libinput *driver_data = pInfo->private;
-	struct libinput_device *device = driver_data->device;
+	struct libinput_device *device = driver_data->shared_device->device;
 	BOOL* data;
 	uint32_t profiles = 0;
 
@@ -1873,7 +2354,7 @@ LibinputSetPropertyNaturalScroll(DeviceIntPtr dev,
 {
 	InputInfoPtr pInfo = dev->public.devicePrivate;
 	struct xf86libinput *driver_data = pInfo->private;
-	struct libinput_device *device = driver_data->device;
+	struct libinput_device *device = driver_data->shared_device->device;
 	BOOL* data;
 
 	if (val->format != 8 || val->size != 1 || val->type != XA_INTEGER)
@@ -1905,7 +2386,7 @@ LibinputSetPropertySendEvents(DeviceIntPtr dev,
 {
 	InputInfoPtr pInfo = dev->public.devicePrivate;
 	struct xf86libinput *driver_data = pInfo->private;
-	struct libinput_device *device = driver_data->device;
+	struct libinput_device *device = driver_data->shared_device->device;
 	BOOL* data;
 	uint32_t modes = 0;
 
@@ -1944,7 +2425,7 @@ LibinputSetPropertyLeftHanded(DeviceIntPtr dev,
 {
 	InputInfoPtr pInfo = dev->public.devicePrivate;
 	struct xf86libinput *driver_data = pInfo->private;
-	struct libinput_device *device = driver_data->device;
+	struct libinput_device *device = driver_data->shared_device->device;
 	BOOL* data;
 
 	if (val->format != 8 || val->size != 1 || val->type != XA_INTEGER)
@@ -1977,7 +2458,7 @@ LibinputSetPropertyScrollMethods(DeviceIntPtr dev,
 {
 	InputInfoPtr pInfo = dev->public.devicePrivate;
 	struct xf86libinput *driver_data = pInfo->private;
-	struct libinput_device *device = driver_data->device;
+	struct libinput_device *device = driver_data->shared_device->device;
 	BOOL* data;
 	uint32_t modes = 0;
 
@@ -2020,7 +2501,7 @@ LibinputSetPropertyScrollButton(DeviceIntPtr dev,
 {
 	InputInfoPtr pInfo = dev->public.devicePrivate;
 	struct xf86libinput *driver_data = pInfo->private;
-	struct libinput_device *device = driver_data->device;
+	struct libinput_device *device = driver_data->shared_device->device;
 	CARD32* data;
 
 	if (val->format != 32 || val->size != 1 || val->type != XA_CARDINAL)
@@ -2054,7 +2535,7 @@ LibinputSetPropertyClickMethod(DeviceIntPtr dev,
 {
 	InputInfoPtr pInfo = dev->public.devicePrivate;
 	struct xf86libinput *driver_data = pInfo->private;
-	struct libinput_device *device = driver_data->device;
+	struct libinput_device *device = driver_data->shared_device->device;
 	BOOL* data;
 	uint32_t modes = 0;
 
@@ -2095,7 +2576,7 @@ LibinputSetPropertyMiddleEmulation(DeviceIntPtr dev,
 {
 	InputInfoPtr pInfo = dev->public.devicePrivate;
 	struct xf86libinput *driver_data = pInfo->private;
-	struct libinput_device *device = driver_data->device;
+	struct libinput_device *device = driver_data->shared_device->device;
 	BOOL* data;
 
 	if (val->format != 8 || val->size != 1 || val->type != XA_INTEGER)
@@ -2126,7 +2607,7 @@ LibinputSetPropertyDisableWhileTyping(DeviceIntPtr dev,
 {
 	InputInfoPtr pInfo = dev->public.devicePrivate;
 	struct xf86libinput *driver_data = pInfo->private;
-	struct libinput_device *device = driver_data->device;
+	struct libinput_device *device = driver_data->shared_device->device;
 	BOOL* data;
 
 	if (val->format != 8 || val->size != 1 || val->type != XA_INTEGER)
@@ -2263,6 +2744,8 @@ LibinputSetProperty(DeviceIntPtr dev, Atom atom, XIPropertyValuePtr val,
 
 	if (atom == prop_tap)
 		rc = LibinputSetPropertyTap(dev, atom, val, checkonly);
+	else if (atom == prop_tap_drag)
+		rc = LibinputSetPropertyTapDrag(dev, atom, val, checkonly);
 	else if (atom == prop_tap_drag_lock)
 		rc = LibinputSetPropertyTapDragLock(dev, atom, val, checkonly);
 	else if (atom == prop_calibration)
@@ -2294,6 +2777,7 @@ LibinputSetProperty(DeviceIntPtr dev, Atom atom, XIPropertyValuePtr val,
 		rc = LibinputSetPropertyHorizScroll(dev, atom, val, checkonly);
 	else if (atom == prop_device || atom == prop_product_id ||
 		 atom == prop_tap_default ||
+		 atom == prop_tap_drag_default ||
 		 atom == prop_tap_drag_lock_default ||
 		 atom == prop_calibration_default ||
 		 atom == prop_accel_default ||
@@ -2367,6 +2851,30 @@ LibinputInitTapProperty(DeviceIntPtr dev,
 						1, &tap);
 }
 
+static void
+LibinputInitTapDragProperty(DeviceIntPtr dev,
+			    struct xf86libinput *driver_data,
+			    struct libinput_device *device)
+{
+	BOOL drag = driver_data->options.tap_drag;
+
+	if (libinput_device_config_tap_get_finger_count(device) == 0)
+		return;
+
+	prop_tap_drag = LibinputMakeProperty(dev,
+					     LIBINPUT_PROP_TAP_DRAG,
+					     XA_INTEGER, 8,
+					     1, &drag);
+	if (!prop_tap_drag)
+		return;
+
+	drag = libinput_device_config_tap_get_default_drag_enabled(device);
+	prop_tap_drag_default = LibinputMakeProperty(dev,
+						     LIBINPUT_PROP_TAP_DRAG_DEFAULT,
+						     XA_INTEGER, 8,
+						     1, &drag);
+}
+
 static void
 LibinputInitTapDragLockProperty(DeviceIntPtr dev,
 				struct xf86libinput *driver_data,
@@ -2384,7 +2892,7 @@ LibinputInitTapDragLockProperty(DeviceIntPtr dev,
 	if (!prop_tap_drag_lock)
 		return;
 
-	drag_lock = libinput_device_config_tap_get_default_enabled(device);
+	drag_lock = libinput_device_config_tap_get_default_drag_lock_enabled(device);
 	prop_tap_drag_lock_default = LibinputMakeProperty(dev,
 							  LIBINPUT_PROP_TAP_DRAG_LOCK_DEFAULT,
 							  XA_INTEGER, 8,
@@ -2835,8 +3343,7 @@ LibinputInitDragLockProperty(DeviceIntPtr dev,
 	size_t sz;
 	int dl_values[MAX_BUTTONS + 1];
 
-	if (!libinput_device_has_capability(driver_data->device,
-					    LIBINPUT_DEVICE_CAP_POINTER))
+	if ((driver_data->capabilities & CAP_POINTER) == 0)
 		return;
 
 	switch (draglock_get_mode(&driver_data->draglock)) {
@@ -2876,7 +3383,7 @@ LibinputInitProperty(DeviceIntPtr dev)
 {
 	InputInfoPtr pInfo  = dev->public.devicePrivate;
 	struct xf86libinput *driver_data = pInfo->private;
-	struct libinput_device *device = driver_data->device;
+	struct libinput_device *device = driver_data->shared_device->device;
 	const char *device_node;
 	CARD32 product[2];
 	int rc;
@@ -2884,6 +3391,7 @@ LibinputInitProperty(DeviceIntPtr dev)
 	prop_float = XIGetKnownProperty("FLOAT");
 
 	LibinputInitTapProperty(dev, driver_data, device);
+	LibinputInitTapDragProperty(dev, driver_data, device);
 	LibinputInitTapDragLockProperty(dev, driver_data, device);
 	LibinputInitCalibrationProperty(dev, driver_data, device);
 	LibinputInitAccelProperty(dev, driver_data, device);
